module PUBMED where

import Control.Applicative
import Data.Attoparsec.ByteString
import Data.Attoparsec.ByteString.Char8 (anyChar)
import Data.ByteString.Char8 (pack)
import Data.ByteString.Lazy (ByteString)
import Data.Text (Text)
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import PUBMED.Client
import PUBMED.Parser
import Prelude hiding (takeWhile)
import Servant.Client (runClientM, mkClientEnv, BaseUrl(..), Scheme(..))
import qualified Data.ByteString.Lazy as LBS
import qualified Data.List as List
import qualified Data.Text as T
import qualified Data.Text.Encoding as TE
import qualified Data.Text.Lazy as TL

pmHost :: String
pmHost = "eutils.ncbi.nlm.nih.gov"
pmSearchPath :: String
pmSearchPath = "entrez/eutils"
pmPort :: Int
pmPort = 443

defaultEnv = do
  manager' <- newManager tlsManagerSettings
  pure $ mkClientEnv manager' $ BaseUrl Https pmHost pmPort pmSearchPath

-- | API main function
getMetadataWith :: Text -> Maybe Limit -> IO (Either Text [PubMed])
getMetadataWith = runSimpleFindPubmedAbstractRequest

-- | TODO this parser need at least one subs at the end
-- (use endOfInput)
removeSub :: Parser ByteString
removeSub = do
  dt <- many textWithBalise
  eo <- manyTill anyChar endOfInput
  pure $ LBS.fromStrict $ pack $ concat dt <> eo
  where
    textWithBalise =
      manyTill anyChar (sub <|> asub)
    sub  = string "<sub>"
      <|> string "<sup>"
      <|> string "<i>"
      <|> string "<b>"
    asub = string "</sub>"
      <|> string "</sup>"
      <|> string "</i>"
      <|> string "</b>"

type Query = Text
type Limit = Integer

runMultipleFPAR :: [Integer]
                -> IO (Either Text [PubMed])
runMultipleFPAR ids = List.foldl1' concat'
                   <$> mapM runSimpleFetchPubmedAbstractRequest (by 300 ids)
  where
    by _ [] = []
    by n ns = head' : ( by n tail')
      where
        (head',tail') = List.splitAt n ns

    concat' (Right n) (Right m) = Right $ n <> m
    concat' n m = n <> m

runSimpleFetchPubmedAbstractRequest ::
  [Integer]
  -> IO (Either Text [PubMed])
runSimpleFetchPubmedAbstractRequest ids = do
  env <- defaultEnv
  res <- runClientM
    (fetch (Just "pubmed") (Just "abstract") ids)
    env
  case res of
    (Left err) -> pure (Left . T.pack $ show err)
    (Right (BsXml abs)) -> do
      --putStrLn $ show abs
      case parseOnly removeSub $ LBS.toStrict abs of
        (Left err'') -> pure (Left $ T.pack err'')
        (Right v) -> do
          let parsed = parsePubMed $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict v
          pure $ Right parsed
--          parsed <- catch (pubMedParser v) ((\e -> do
--                                                _ <- print e
--                                                pure []) :: XmlException -> IO [PubMed])
--          Right <$> pure parsed

runSimpleFindPubmedAbstractRequest :: Text -> Maybe Limit -> IO (Either Text [PubMed])
runSimpleFindPubmedAbstractRequest query limit =  do
  eDocIds <- searchDocIds query limit
  case eDocIds of
    Left err -> pure $ Left err
    Right docIds -> runMultipleFPAR docIds

searchDocIds :: Text -> Maybe Limit -> IO (Either Text [Integer])
searchDocIds query limit = do
  env <- defaultEnv
  res <- runClientM
    (search (Just query) limit)
    env
  case res of
    (Left err) -> pure (Left $ T.pack $ show err)
    (Right (BsXml docs)) -> do
      pure $ Right $ parseDocIds $ TL.fromStrict $ TE.decodeUtf8 $ LBS.toStrict docs
      --pure $ Right $ runParser parseDocId docs

--runParser :: Show res => (Cursor -> res) -> LBS.ByteString -> res
--runParser parser = parser . fromDocument . parseLBS_ def
